/*************************************************************************
* Copyright (c) 2008 High Energy Accelerator Reseach Organization (KEK)
*
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
**************************************************************************
* drvF3RP61.c - Driver Support Routines for F3RP61
*
*      Author: Jun-ichi Odagiri 
*      Date: 6-30-08
*/

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/msg.h>
#include <fcntl.h>
#include <asm/fam3rtos/m3iodrv.h>

#include <dbCommon.h>
#include <dbScan.h>
#include <recSup.h>
#include <drvSup.h>
#include <iocsh.h>
#include <epicsThread.h>
#ifndef EPICS_REVISION
#include <epicsVersion.h>
#endif
#include <epicsExport.h>
#include "drvF3RP61.h"

static long report();
static long init();

struct
{
  long      number;
  DRVSUPFUN report;
  DRVSUPFUN init;
} drvF3RP61 =
{
  2L,
  report,
  init,
};

epicsExportAddress(drvet,drvF3RP61);

int f3rp61_fd;
F3RP61_IO_INTR f3rp61_io_intr[M3IO_NUM_UNIT][M3IO_NUM_SLOT];
M3IO_MODULE_INFORMATION f3rp61_module_info[M3IO_NUM_UNIT][M3IO_NUM_SLOT];
long f3rp61_register_io_interrupt(dbCommon *, int, int, int);

static int f3rp61_init_flag;
static void msgrcv_thread(void *);


static long report(void)
{

  return (0);
}


static long init(void)
{
  int i, j;
  
  printf("drvF3RP61: init() entered\n");

  if (f3rp61_init_flag) return (0);
  f3rp61_init_flag = 1;

  f3rp61_fd = open("/dev/m3io", O_RDWR);
  if (f3rp61_fd < 0) {
    errlogPrintf("drvF3RP61: can't open /dev/m3io\n");
    return (-1);
  }

  for (i = 0; i < M3IO_NUM_UNIT; i++) {
    for (j = 0; j < M3IO_NUM_SLOT; j++) {
      f3rp61_module_info[i][j].unitno = i;
      f3rp61_module_info[i][j].slotno = j;
      ioctl(f3rp61_fd, M3IO_GET_MODULE_INFO, &f3rp61_module_info[i][j]);

    }
  }

  return (0);
}


static void msgrcv_thread(void *arg)
{
  int msqid = (int) arg;
  MSG_BUF msgbuf;
  int unit;
  int slot;
  int channel;
  dbCommon *prec;
  IOSCANPVT ioscanpvt;
  int i;
  for (;;) {
    if (msgrcv(msqid, &msgbuf, sizeof(MSG_BUF), M3IO_MSGTYPE_IO, 0) == -1) {
      errlogPrintf("drvF3RP61: msgrcv failed\n");
    }
    unit = msgbuf.mtext.unit;
    slot = msgbuf.mtext.slot;
    channel = msgbuf.mtext.channel;
    printf("unit=%d,slot=%d,channel=%d\n",unit,slot,channel);

    for (i = 0; i < f3rp61_io_intr[unit][slot].count; i++) {
      if (f3rp61_io_intr[unit][slot].ioscan[i].channel == channel) {
	prec = f3rp61_io_intr[unit][slot].ioscan[i].prec;
	if (!prec) {
	  errlogPrintf("drvF3RP61: no record for interrupt (U%d,S%d,C%d)\n", unit, slot, channel);
	  break;
	}
	ioscanpvt = *((IOSCANPVT *) prec->dpvt);
	if (prec->scan == SCAN_IO_EVENT) {
	  scanIoRequest(ioscanpvt);
	}
      }
    }
  }
}


long f3rp61_register_io_interrupt(dbCommon *prec, int unit, int slot, int channel)
{
  M3IO_INTER_DEFINE arg;
  char thread_name[32];
  int msqid;
  int count = f3rp61_io_intr[unit][slot].count;
  int i;

  /*
  printf("drvF3RP61: f3rp61_register_io_interrupt() entered for %d\n", count);
  */
  if (count == NUM_IO_INTR) {
    errlogPrintf("drvF3RP61: no interrupt slot\n");
    return (-1);
  }

  if (count == 0) {
    if ((msqid = msgget(IPC_PRIVATE, IPC_CREAT | 0666)) == -1) {
      errlogPrintf("drvF3RP61: msgget failed\n");
      return (-1);
    }
  /* Add Start */
    if ((msqid = msgget(IPC_PRIVATE, IPC_CREAT | 0666)) == -1) {
      errlogPrintf("drvF3RP61: msgget failed\n");
      return (-1);
    }
  /* Add End */

    sprintf(thread_name, "msgrcvr%d", msqid);
    if ((epicsThreadCreate(thread_name,
			   epicsThreadPriorityHigh,
			   epicsThreadGetStackSize(epicsThreadStackSmall),
			   (EPICSTHREADFUNC) msgrcv_thread,
			   (void *) msqid)) == (epicsThreadId) 0) {
      errlogPrintf("drvF3RP61: epicsThreadCreate failed\n");
      return (-1);
    }
  }

  f3rp61_io_intr[unit][slot].ioscan[count].channel = channel;
  f3rp61_io_intr[unit][slot].ioscan[count].prec = prec;
  f3rp61_io_intr[unit][slot].count++;
  /* scanIoInit((IOSCANPVT *) prec->dpvt); */

  for (i = 0; i < count; i++) {
    if (f3rp61_io_intr[unit][slot].ioscan[i].channel == channel) {
      return (0);
    }
  }

  arg.unitno = (unsigned short) unit;
  arg.slotno = (unsigned short) slot;
  arg.defData.relayNo = (long) channel;
  arg.msgQId = msqid;

  if (ioctl(f3rp61_fd, M3IO_ENABLE_INTER, &arg) < 0) {
    errlogPrintf("devBiF3RP61: ioctl failed\n");		        
    return (-1);
  }

  return (0);
}


/*******************************************************************************
 * Get io interrupt info
 *******************************************************************************/
long f3rp61GetIoIntInfo(int cmd, struct dbCommon * pxx, IOSCANPVT *ppvt)
{
  if (!pxx->dpvt) {
    errlogPrintf("drvF3RP61: f3rp61GetIoIntInfo is called with null dpvt\n");
    return (-1);
  }

  if ( *((IOSCANPVT *) pxx->dpvt) == NULL) {
    scanIoInit((IOSCANPVT *) pxx->dpvt);
  }

  *ppvt = *((IOSCANPVT *) pxx->dpvt);

  return (0);
}
